from typing import Any, Optional

from .site import site_metadata_rules


class MetadataRuleEvaluator:

    """Main class to evaluate rules.

    Used by model mixin.
    """

    def __init__(
        self, related_visit: Optional[Any] = None, app_label: Optional[str] = None
    ) -> None:
        self.related_visit = related_visit
        self.app_label = app_label or related_visit._meta.app_label

    def evaluate_rules(self) -> None:
        for rule_group in site_metadata_rules.registry.get(self.app_label, []):
            rule_group.evaluate_rules(visit=self.related_visit)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..metadata_updater import MetadataUpdater
from .requisition_target_handler import RequisitionTargetHandler

if TYPE_CHECKING:
    from edc_lab.models import Panel


class RequisitionMetadataError(Exception):
    pass


class RequisitionMetadataUpdater(MetadataUpdater):

    """A class to update a subject's requisition metadata given
    the visit, target model name, panel and desired entry status.
    """

    target_handler: RequisitionTargetHandler = RequisitionTargetHandler

    def __init__(self, target_panel: Panel = None, **kwargs):
        super().__init__(**kwargs)
        self.target_panel = target_panel

    @property
    def target(self):
        target = self.target_handler(
            model=self.target_model,
            related_visit=self.related_visit,
            target_panel=self.target_panel,
        )
        return target



from collections import OrderedDict
from typing import Any, Optional

from ...constants import NOT_REQUIRED, REQUIRED
from ...metadata_updater import MetadataUpdater
from ...target_handler import TargetModelConflict
from ..rule_group import RuleGroup
from ..rule_group_metaclass import RuleGroupMetaclass


class CrfRuleGroup(RuleGroup, metaclass=RuleGroupMetaclass):
    """A class used to declare and contain rules."""

    metadata_updater_cls = MetadataUpdater

    def __str__(self: Any):
        return f"{self.__class__.__name__}({self.name})"

    def __repr__(self: Any):
        return f"{self.__class__.__name__}({self.name})"

    @classmethod
    def crfs_for_visit(cls, visit=None):
        """Returns a list of scheduled or unscheduled
        CRFs + PRNs depending on visit_code_sequence.
        """
        if visit.visit_code_sequence != 0:
            crfs = visit.visit.crfs_unscheduled + visit.visit.crfs_prn
        else:
            crfs = visit.visit.crfs + visit.visit.crfs_prn
        return crfs

    @classmethod
    def evaluate_rules(cls: Any, visit: Any = None) -> tuple:
        rule_results = OrderedDict()
        metadata_objects = OrderedDict()
        for rule in cls._meta.options.get("rules"):
            # skip if source model is not in visit.crfs (including PRNs)
            if (
                rule.source_model
                and rule.source_model != visit._meta.label_lower
                and rule.source_model not in [c.model for c in cls.crfs_for_visit(visit)]
            ):
                continue
            rule_results.update({str(rule): rule.run(visit=visit)})
            for target_model, entry_status in rule_results[str(rule)].items():
                if target_model == visit._meta.label_lower:
                    raise TargetModelConflict(
                        f"Target model and visit model are the same! "
                        f"Got {target_model}=={visit._meta.label_lower}"
                    )
                # only do something if target model is in visit.crfs (including PRNs)
                if target_model in [c.model for c in cls.crfs_for_visit(visit)]:
                    metadata_updater = cls.metadata_updater_cls(
                        related_visit=visit, target_model=target_model
                    )
                    metadata_obj = metadata_updater.update(
                        entry_status=entry_status
                        or cls.default_entry_status(visit, target_model)
                    )
                    metadata_objects.update({target_model: metadata_obj})
        return rule_results, metadata_objects

    @classmethod
    def default_entry_status(cls, visit: Any, target_model: Any) -> Optional[str]:
        """Returns the default `entry_status` or None"""
        for c in cls.crfs_for_visit(visit):
            if c.model == target_model:
                return REQUIRED if c.required else NOT_REQUIRED
        return None



from typing import Any, Optional

from ...constants import REQUISITION
from ..rule import Rule
from ..rule_evaluator import RuleEvaluator


class RequisitionRuleEvaluatorError(Exception):
    pass


class RequisitionRuleEvaluator(RuleEvaluator):
    def __init__(self, source_panel: Optional[Any] = None, **kwargs) -> None:
        self.source_panel = source_panel
        super().__init__(**kwargs)


class RequisitionRule(Rule):

    rule_evaluator_cls = RequisitionRuleEvaluator

    def __init__(
        self,
        source_panel: Optional[Any] = None,
        target_panels: Optional[list] = None,
        **kwargs,
    ) -> None:
        self.metadata_category = REQUISITION
        self.target_panels = [p for p in target_panels]
        self.source_panel = source_panel
        super().__init__(**kwargs)



from .updates_crf_metadata_model_mixin import UpdatesCrfMetadataModelMixin
from .updates_metadata_model_mixin import UpdatesMetadataModelMixin
from .updates_requisition_metadata_model_mixin import (
    UpdatesRequisitionMetadataModelMixin,
)



from typing import Any

from .crf import CrfRuleGroup
from .requisition import RequisitionRuleGroup
from .site import site_metadata_rules


class RegisterRuleGroupError(Exception):
    pass


def register(site=None, **kwargs) -> Any:  # noqa
    """Registers a rule group."""
    site = site or site_metadata_rules

    def _rule_group_wrapper(rule_group_cls: Any) -> Any:

        if not issubclass(rule_group_cls, (CrfRuleGroup, RequisitionRuleGroup)):
            raise RegisterRuleGroupError(
                f"Wrapped class must a RuleGroup class. Got {rule_group_cls}"
            )

        site.register(rule_group_cls=rule_group_cls)

        return rule_group_cls

    return _rule_group_wrapper



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist

if TYPE_CHECKING:
    from edc_crf.model_mixins import CrfModelMixin
    from edc_visit_tracking.model_mixins import VisitModelMixin

    from edc_metadata.models import CrfMetadata, RequisitionMetadata


class MetadataWrapperError(Exception):
    pass


class MetadataWrapper:

    """A class that wraps the corresponding model instance, or not, for the
    given metadata object and sets it to itself along with other
    attributes like the visit, model class, metadata_obj, etc.
    """

    label: str | None = None

    def __init__(
        self,
        visit: VisitModelMixin,
        metadata_obj: CrfMetadata | RequisitionMetadata,
    ) -> None:
        self._model_obj = None
        self.metadata_obj = metadata_obj
        self.visit = visit

        # visit codes (and sequence) must match
        if (self.visit.visit_code != self.metadata_obj.visit_code) or (
            self.visit.visit_code_sequence != self.metadata_obj.visit_code_sequence
        ):
            raise MetadataWrapperError(
                f"Visit code mismatch. Visit is {self.visit.visit_code}."
                f"{self.visit.visit_code_sequence} but metadata object has "
                f"{self.metadata_obj.visit_code}."
                f"{self.metadata_obj.visit_code_sequence}. Got {repr(metadata_obj)}."
            )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.visit}, {self.metadata_obj})"

    @property
    def options(self) -> dict:
        """Returns a dictionary of query options."""
        return {f"{self.model_cls.related_visit_model_attr()}": self.visit}

    @property
    def model_obj(self) -> CrfModelMixin:
        if not self._model_obj:
            try:
                self._model_obj = self.model_cls.objects.get(**self.options)
            except AttributeError as e:
                if "related_visit_model_attr" not in str(e):
                    raise ImproperlyConfigured(f"{e} See {repr(self.model_cls)}")
            except ObjectDoesNotExist:
                self._model_obj = None
        return self._model_obj

    @model_obj.setter
    def model_obj(self, value=None) -> None:
        self._model_obj = value

    @property
    def model_cls(self) -> CrfModelMixin:
        """Returns a CRF model class"""
        return django_apps.get_model(self.metadata_obj.model)



from model_bakery import baker

from ..constants import NOT_REQUIRED, REQUIRED
from ..models import CrfMetadata, RequisitionMetadata


class CrfTestHelper:

    """A test class to help create CRFs and manipulate
    metadata.
    """

    def baker_options(self, report_datetime):
        """Override the default recipe options for your
        'baker recipe {label_lower: {key: value}, ...}.
        """
        return {}

    def crf_metadata_obj(self, model, entry_status, visit_code):
        return CrfMetadata.objects.filter(
            entry_status=entry_status,
            model=model,
            visit_code=visit_code,
            subject_identifier=self.subject_identifier,
        )

    def requisition_metadata_obj(self, model, entry_status, visit_code, panel_name):
        return RequisitionMetadata.objects.filter(
            entry_status=entry_status,
            model=model,
            subject_identifier=self.subject_identifier,
            panel_name=panel_name,
            visit_code=visit_code,
        )

    def get_crfs(self, visit_code=None, subject_identifier=None):
        """Return a queryset of crf metadata for the visit."""
        return CrfMetadata.objects.filter(
            subject_identifier=subject_identifier, visit_code=visit_code
        ).order_by("show_order")

    def get_crfs_by_entry_status(
        self, visit_code=None, entry_status=None, subject_identifier=None
    ):
        """Return a queryset of crf metadata for the visit by entry_status."""
        return (
            self.get_crfs(visit_code=visit_code, subject_identifier=subject_identifier)
            .filter(entry_status__in=entry_status)
            .order_by("show_order")
        )

    def complete_crfs(
        self,
        visit_code=None,
        visit=None,
        visit_attr=None,
        entry_status=None,
        subject_identifier=None,
    ):
        """Complete all CRFs in a visit by looping through metadata.

        Revisit the metadata on each loop as rule_groups may change
        the entry status of CRFs.
        """
        entry_status = entry_status or [REQUIRED, NOT_REQUIRED]
        visit_attr = visit_attr or "subject_visit"
        if not isinstance(entry_status, (list, tuple)):
            entry_status = [entry_status]
        completed_crfs = []
        while True:
            for crf in self.get_crfs_by_entry_status(
                visit_code=visit_code,
                entry_status=entry_status,
                subject_identifier=subject_identifier,
            ):
                options = self.baker_options(visit.report_datetime).get(crf.model, {})
                options.update({visit_attr: visit, "report_datetime": visit.report_datetime})
                completed_crfs.append(baker.make_recipe(crf.model, **options))
            if not self.get_crfs_by_entry_status(
                visit_code=visit_code,
                entry_status=entry_status,
                subject_identifier=subject_identifier,
            ):
                break
        return completed_crfs

    def complete_required_crfs(
        self, visit_code=None, visit=None, visit_attr=None, subject_identifier=None
    ):
        return self.complete_crfs(
            visit_code=visit_code,
            visit=visit,
            visit_attr=visit_attr,
            entry_status=REQUIRED,
            subject_identifier=subject_identifier,
        )



from collections import OrderedDict

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase
from edc_constants.constants import MALE
from edc_facility.import_holidays import import_holidays
from edc_reference.site_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from faker import Faker

from edc_metadata import KEYED, NOT_REQUIRED, REQUIRED
from edc_metadata.models import CrfMetadata

from ...metadata_rules import CrfRule, CrfRuleGroup, P, site_metadata_rules
from ..models import Appointment, CrfOne, CrfTwo, SubjectConsent, SubjectVisit
from ..reference_configs import register_to_site_reference_configs
from ..visit_schedule import visit_schedule

fake = Faker()
edc_registration_app_config = django_apps.get_app_config("edc_registration")


class CrfRuleGroupOne(CrfRuleGroup):

    crfs_car = CrfRule(
        predicate=P("f1", "eq", "car"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crftwo"],
    )

    crfs_bicycle = CrfRule(
        predicate=P("f3", "eq", "bicycle"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfthree"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_metadata.crfone"


class CrfRuleGroupTwo(CrfRuleGroup):

    crfs_truck = CrfRule(
        predicate=P("f1", "eq", "truck"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crffive"],
    )

    crfs_train = CrfRule(
        predicate=P("f1", "eq", "train"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfsix"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_metadata.crfone"


class CrfRuleGroupThree(CrfRuleGroup):

    crfs_truck = CrfRule(
        predicate=P("f1", "eq", "holden"),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["prnone"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_metadata.crfone"


class CrfRuleGroupTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        register_to_site_reference_configs()
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_metadata.subjectvisit"}
        )

        # note crfs in visit schedule are all set to REQUIRED by default.
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )

        site_metadata_rules.registry = OrderedDict()
        site_metadata_rules.register(rule_group_cls=CrfRuleGroupOne)
        site_metadata_rules.register(rule_group_cls=CrfRuleGroupTwo)
        site_metadata_rules.register(rule_group_cls=CrfRuleGroupThree)

    def enroll(self, gender=None):
        subject_identifier = fake.credit_card_number()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=subject_identifier,
            consent_datetime=get_utcnow(),
            gender=gender,
        )
        self.schedule.put_on_schedule(
            subject_identifier=subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            reason=SCHEDULED,
            subject_identifier=subject_identifier,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
        )
        return subject_visit

    def get_next_subject_visit(self, subject_visit):
        return SubjectVisit.objects.create(
            appointment=self.appointment.next,
            reason=SCHEDULED,
            subject_identifier=subject_visit.subject_identifier,
            visit_schedule_name=self.appointment.next.visit_schedule_name,
            schedule_name=self.appointment.next.schedule_name,
            visit_code=self.appointment.next.visit_code,
            visit_code_sequence=self.appointment.next.visit_code_sequence,
        )

    def test_example1(self):
        """Asserts CrfTwo is REQUIRED if f1==\'car\' as specified."""
        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            REQUIRED,
        )

        CrfOne.objects.create(subject_visit=subject_visit, f1="car")

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            NOT_REQUIRED,
        )

    def test_example2(self):
        """Asserts CrfThree is REQUIRED if f1==\'bicycle\' as specified."""

        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            REQUIRED,
        )

        CrfOne.objects.create(subject_visit=subject_visit, f3="bicycle")

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            NOT_REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            REQUIRED,
        )

        subject_visit.save()

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            NOT_REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            REQUIRED,
        )

    def test_example4(self):
        """Asserts CrfThree is REQUIRED if f1==\'bicycle\' but then not
        when f1 is changed to \'car\' as specified
        by edc_example.rule_groups.ExampleRuleGroup2."""

        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            REQUIRED,
        )

        crf_one = CrfOne.objects.create(subject_visit=subject_visit, f1="not car")

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            NOT_REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            NOT_REQUIRED,
        )

        crf_one.f1 = "car"
        crf_one.save()

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            NOT_REQUIRED,
        )

        crf_one.f3 = "bicycle"
        crf_one.save()

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crfthree").entry_status,
            REQUIRED,
        )

    def test_keyed_instance_ignores_rules(self):
        """Asserts if instance exists, rule is ignored."""
        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )

        CrfTwo.objects.create(subject_visit=subject_visit)

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            KEYED,
        )

        crf_one = CrfOne.objects.create(subject_visit=subject_visit, f1="not car")

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            KEYED,
        )

        crf_one.f1 = "car"
        crf_one.save()

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            KEYED,
        )

    def test_recovers_from_missing_metadata(self):
        subject_visit = self.enroll(gender=MALE)
        metadata_obj = CrfMetadata.objects.get(model="edc_metadata.crftwo")
        self.assertEqual(metadata_obj.entry_status, REQUIRED)

        metadata_obj.delete()

        CrfTwo.objects.create(subject_visit=subject_visit)

        metadata_obj = CrfMetadata.objects.get(model="edc_metadata.crftwo")
        self.assertEqual(metadata_obj.entry_status, KEYED)

    def test_delete(self):
        """Asserts delete returns to default entry status."""
        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )

        crf_two = CrfTwo.objects.create(subject_visit=subject_visit)

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            KEYED,
        )

        crf_two.delete()

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )

    def test_delete_2(self):
        """Asserts delete returns to entry status of rule for crf_two."""
        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            REQUIRED,
        )

        crf_two = CrfTwo.objects.create(subject_visit=subject_visit)

        CrfOne.objects.create(subject_visit=subject_visit, f1="not car")

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            KEYED,
        )

        crf_two.delete()

        self.assertEqual(
            CrfMetadata.objects.get(model="edc_metadata.crftwo").entry_status,
            NOT_REQUIRED,
        )

    def test_prn_is_created_as_not_required_by_default(self):
        """Asserts handles PRNs correctly"""
        # create both visits before going back to add crf_one
        subject_visit = self.enroll(gender=MALE)
        self.assertEqual(
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit.visit_code,
                visit_code_sequence=subject_visit.visit_code_sequence,
            ).entry_status,
            NOT_REQUIRED,
        )

    def test_prn_rule_acts_on_correct_visit(self):
        """Asserts handles PRNs correctly"""
        # create both visits before going back to add crf_one
        subject_visit = self.enroll(gender=MALE)
        subject_visit_two = self.get_next_subject_visit(subject_visit)
        crf_one = CrfOne.objects.create(subject_visit=subject_visit, f1="holden")
        self.assertEqual(
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit.visit_code,
                visit_code_sequence=subject_visit.visit_code_sequence,
            ).entry_status,
            REQUIRED,
        )
        try:
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit_two.visit_code,
                visit_code_sequence=subject_visit_two.visit_code_sequence,
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.fail("CrfMetadata model instance unexpectedly exists")

        crf_one.f1 = "caufield"
        crf_one.save()
        self.assertEqual(
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit.visit_code,
                visit_code_sequence=subject_visit.visit_code_sequence,
            ).entry_status,
            NOT_REQUIRED,
        )
        try:
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit_two.visit_code,
                visit_code_sequence=subject_visit_two.visit_code_sequence,
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.fail("CrfMetadata model instance unexpectedly exists")
        crf_one.f1 = "holden"
        crf_one.save()
        self.assertEqual(
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit.visit_code,
                visit_code_sequence=subject_visit.visit_code_sequence,
            ).entry_status,
            REQUIRED,
        )
        try:
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit_two.visit_code,
                visit_code_sequence=subject_visit_two.visit_code_sequence,
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.fail("CrfMetadata model instance unexpectedly exists")

    def test_prn_rule_acts_on_correct_visit2(self):
        """Asserts handles PRNs correctly"""
        # create 1000 then add crf_one then create 2000
        subject_visit = self.enroll(gender=MALE)
        crf_one = CrfOne.objects.create(subject_visit=subject_visit, f1="caufield")
        self.assertEqual(1, CrfMetadata.objects.filter(model="edc_metadata.prnone").count())
        self.assertEqual(
            1,
            CrfMetadata.objects.filter(
                model="edc_metadata.prnone", entry_status=NOT_REQUIRED
            ).count(),
        )

        self.get_next_subject_visit(subject_visit)

        self.assertEqual(1, CrfMetadata.objects.filter(model="edc_metadata.prnone").count())
        self.assertEqual(
            1,
            CrfMetadata.objects.filter(
                model="edc_metadata.prnone", entry_status=NOT_REQUIRED
            ).count(),
        )
        self.assertEqual(
            0,
            CrfMetadata.objects.filter(
                model="edc_metadata.prnone", entry_status=REQUIRED
            ).count(),
        )

        crf_one.f1 = "holden"
        crf_one.save()
        self.assertEqual(
            1,
            CrfMetadata.objects.filter(
                model="edc_metadata.prnone", entry_status=REQUIRED
            ).count(),
        )
        self.assertEqual(
            0,
            CrfMetadata.objects.filter(
                model="edc_metadata.prnone", entry_status=NOT_REQUIRED
            ).count(),
        )

    def test_prn_resets_on_delete(self):
        subject_visit = self.enroll(gender=MALE)
        crf_one = CrfOne.objects.create(subject_visit=subject_visit, f1="holden")
        crf_one.delete()
        self.assertEqual(
            CrfMetadata.objects.get(
                model="edc_metadata.prnone",
                visit_code=subject_visit.visit_code,
                visit_code_sequence=subject_visit.visit_code_sequence,
            ).entry_status,
            NOT_REQUIRED,
        )



import sys
from typing import Any, Optional

from django.apps import apps as django_apps
from django.contrib.admin.sites import all_sites
from django.db.models import Count
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.utils import get_subject_visit_model_cls
from tqdm import tqdm

from edc_metadata import InvalidTargetPanel, TargetModelNotScheduledForVisit
from edc_metadata.models import CrfMetadata, RequisitionMetadata

from .metadata_rules import site_metadata_rules


class MetadataRefresher:
    """A class to be `run` when metadata gets out-of-date

    This may happen when there are changes to the visit schedule,
    metada rules or manual changes to data.
    """

    def __init__(self, verbose: Optional[bool] = None):
        self._source_models = []
        self.verbose = verbose

    def run(self) -> None:
        self._message("Updating references, updating metadata ...\n")
        for index, source_model in enumerate(self.source_models):
            self._message(f"  {index + 1}. {source_model}\n")
            source_model_cls = django_apps.get_model(source_model)
            total = source_model_cls.objects.all().count()
            self.update_references(source_model_cls, total)
            self._message("    - Updating metadata ...     \n")
            self.create_or_update_metadata(source_model_cls, total)
        total = get_subject_visit_model_cls().objects.all().count()
        self._message("Running metadata rules ...\n")
        self.run_metadata_rules(get_subject_visit_model_cls(), total)
        self._message("Done.\n")

    @property
    def source_models(self) -> list:
        if not self._source_models:
            self._source_models = [get_subject_visit_model_cls()._meta.label_lower]
            for app_label, rule_groups_list in site_metadata_rules.rule_groups.items():
                for rule_groups in rule_groups_list:
                    self._source_models.append(rule_groups._meta.source_model)
            self._source_models = list(set(self._source_models))
            self._message(f"  Found source models: {', '.join(self.source_models)}.\n")
        return self._source_models

    def update_references(self, source_model_cls: Any, total: int) -> None:
        """Updates references for all instances of this source model"""
        self._message("    - Updating references ...\n")
        for instance in tqdm(source_model_cls.objects.all(), total=total):
            instance.update_reference_on_save()

    def run_metadata_rules(self, source_model_cls: Any, total: int) -> None:
        """Updates rules for all instances of this source model"""
        self._message("    - Running rules ...     \n")
        for instance in tqdm(source_model_cls.objects.all(), total=total):
            if django_apps.get_app_config("edc_metadata").metadata_rules_enabled:
                instance.run_metadata_rules()

    def create_or_update_metadata(self, source_model_cls: Any, total: int) -> None:
        """Creates or updates CRF/Requisition metadata for all instances
        of this source model.
        """
        for instance in tqdm(source_model_cls.objects.all(), total=total):
            try:
                instance.metadata_create()
            except AttributeError:
                try:
                    instance.metadata_update()
                except (
                    AttributeError,
                    TargetModelNotScheduledForVisit,
                    InvalidTargetPanel,
                ) as e:
                    self._message(f"      skipping (got {e})     \n")
                    break

    def create_or_update_metadata_for_all(self) -> None:
        admin_models = []
        for admin_site in all_sites:
            admin_models.extend([cls._meta.label_lower for cls in admin_site._registry])
        crf_metadata_models = [
            obj.get("model")
            for obj in CrfMetadata.objects.values("model")
            .order_by("model")
            .annotate(count=Count("model"))
        ]
        requisition_metadata_models = [
            obj.get("model")
            for obj in RequisitionMetadata.objects.values("model")
            .order_by("model")
            .annotate(count=Count("model"))
        ]
        self._message("    - Verifying CrfMetadata models with visit schedule and admin.\n")
        for model in crf_metadata_models:
            if (
                model not in site_visit_schedules.all_post_consent_models
                and model not in admin_models
            ):
                count = CrfMetadata.objects.filter(model=model).delete()
                self._message(f"      * deleted {count} metadata records for model {model}.\n")
        self._message(
            "    - Verifying RequisitionMetadata models with visit schedule and admin.\n"
        )
        for model in requisition_metadata_models:
            if (
                model not in site_visit_schedules.all_post_consent_models
                and model not in admin_models
            ):
                count = RequisitionMetadata.objects.filter(model=model).delete()
                self._message(f"      * deleted {count} metadata records for model {model}.\n")

        self._message("    - Updating metadata ...     \n")
        for model in site_visit_schedules.all_post_consent_models:
            model_cls = django_apps.get_model(model)
            total = model_cls.objects.all().count()
            self._message(f"      * {model_cls._meta.label_lower} ...     \n")
            self.create_or_update_metadata(model_cls, total)

    def _message(self, msg: str) -> None:
        if self.verbose:
            sys.stdout.write(msg)



from .requisition_rule import (
    RequisitionRule,
    RequisitionRuleEvaluator,
    RequisitionRuleEvaluatorError,
)
from .requisition_rule_group import (
    RequisitionRuleGroup,
    RequisitionRuleGroupMetaOptions,
    RequisitionRuleGroupMetaOptionsError,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from .metadata_wrapper import MetadataWrapper

if TYPE_CHECKING:
    from ..models import RequisitionMetadata


class RequisitionMetadataWrapper(MetadataWrapper):

    label = "Requisition"

    def __init__(self, metadata_obj: RequisitionMetadata = None, **kwargs) -> None:
        self.panel_name = metadata_obj.panel_name
        super().__init__(metadata_obj=metadata_obj, **kwargs)

    @property
    def options(self) -> dict:
        options = super().options
        options.update(panel__name=self.panel_name)
        return options

    @property
    def html_id(self) -> str:
        return f"id_{self.panel_name}"



import copy
import sys
from collections import OrderedDict
from typing import Any, Optional

from django.apps import apps as django_apps
from django.utils.module_loading import import_module, module_has_submodule


class SiteMetadataRulesAlreadyRegistered(Exception):
    pass


class SiteMetadataNoRulesError(Exception):
    pass


class SiteMetadataRules:

    """Main controller of :class:`MetadataRules` objects."""

    def __init__(self) -> None:
        self.registry = OrderedDict()

    def register(self, rule_group_cls: Optional[Any] = None) -> None:
        """Register MetadataRules to a list per app_label
        for the module the rule groups were declared in.
        """
        if rule_group_cls:
            if not rule_group_cls._meta.options.get("rules"):
                raise SiteMetadataNoRulesError(
                    f"The metadata rule group {rule_group_cls.name} " f"has no rule!"
                )

            if rule_group_cls._meta.app_label not in self.registry:
                self.registry.update({rule_group_cls._meta.app_label: []})

            for rgroup in self.registry.get(rule_group_cls._meta.app_label):
                if rgroup.name == rule_group_cls.name:
                    raise SiteMetadataRulesAlreadyRegistered(
                        f"The metadata rule group {rule_group_cls.name} "
                        f"is already registered"
                    )
            self.registry.get(rule_group_cls._meta.app_label).append(rule_group_cls)

    @property
    def rule_groups(self) -> Any:
        return self.registry

    def validate(self) -> None:
        for rule_groups in self.registry.values():
            for rule_group in rule_groups:
                sys.stdout.write(f"{repr(rule_group)}\n")
                rule_group.validate()

    @staticmethod
    def autodiscover(module_name: Optional[str] = None) -> None:
        """Autodiscovers rules in the metadata_rules.py file
        of any INSTALLED_APP.
        """
        module_name = module_name or "metadata_rules"
        sys.stdout.write(f" * checking for {module_name} ...\n")
        for app in django_apps.app_configs:
            try:
                before_import_registry = None
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_metadata_rules.registry)
                    import_module(f"{app}.{module_name}")
                except Exception as e:
                    if f"No module named '{app}.{module_name}'" not in str(e):
                        site_metadata_rules.registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise

                else:
                    sys.stdout.write(
                        f"   - imported metadata rules from '{app}.{module_name}'\n"
                    )
            except ImportError:
                pass


site_metadata_rules = SiteMetadataRules()



from .requisition_metadata_updater import (
    RequisitionMetadataError,
    RequisitionMetadataUpdater,
)
from .requisition_target_handler import (
    InvalidTargetPanel,
    RequisitionTargetHandler,
    TargetPanelNotScheduledForVisit,
)



from collections import OrderedDict
from typing import Any, Optional

from edc_appointment.constants import MISSED_APPT

from .logic import Logic
from .rule_evaluator import RuleEvaluator


class RuleError(Exception):
    pass


class Rule:

    rule_evaluator_cls = RuleEvaluator
    logic_cls = Logic

    def __init__(
        self,
        predicate: Optional[Any] = None,
        consequence: Optional[str] = None,
        alternative: Optional[str] = None,
    ) -> None:
        self._logic = self.logic_cls(
            predicate=predicate, consequence=consequence, alternative=alternative
        )
        self.target_models = None
        self.app_label = None  # set by metaclass
        self.group = None  # set by metaclass
        self.name = None  # set by metaclass
        self.source_model = None  # set by metaclass
        self.reference_getter_cls = None  # set by metaclass
        self.field_names = []
        try:
            self.field_names = [predicate.attr]
        except AttributeError:
            try:
                self.field_names = predicate.attrs
            except AttributeError:
                pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name='{self.name}', group='{self.group}')"

    def __str__(self) -> str:
        return f"{self.group}.{self.name}"

    def run(self, visit: Optional[Any] = None) -> dict:
        """Returns a dictionary of {target_model: entry_status, ...} updated
        by running the rule for each target model given a visit.

        Skips run if `appointment.appt_timing` == MISSED_APPT

        Ensure the `model.field` is registered with `site_reference_configs`.
        See `edc_reference`.
        """
        result = OrderedDict()
        if visit.appointment.appt_timing != MISSED_APPT:
            opts = {k: v for k, v in self.__dict__.items() if k.startswith != "_"}
            rule_evaluator = self.rule_evaluator_cls(visit=visit, logic=self._logic, **opts)
            entry_status = rule_evaluator.result
            for target_model in self.target_models:
                result.update({target_model: entry_status})
        return result
